/**----------------------------------------------------------------------------------------------------
 * easy namespace
 * 
 * Dependencies: node-walk, backbone
 *-----------------------------------------------------------------------------------------------------
 */

// External libraries
var Backbone    = require('backbone');
var _           = require('underscore');
var walk        = require('walk');

// Node libraries
var path        = require('path');
var fs          = require('fs');

//-------------------------------------------------------------------------------
// Converts a dash-delimited string into a camelCased one
function toCamel(str) {
    return str.replace(/(\-[a-z])/g, function($1){return $1.toUpperCase().replace('-','');});
};

//-------------------------------------------------------------------------------
// Converts a dash-delimited string into a PascalCased one
function toPascal(str) {
    str = str.replace(/(\-[a-z])/g, function($1){return $1.toUpperCase().replace('-','');});
    return str.charAt(0).toUpperCase() + str.slice(1);
};

//-------------------------------------------------------------------------------
// Counter
var _loaded = 0;

//======================================================================================================
// Easy namespace
var easy = module.exports = {
    
    //---------------------------------------------------------------------------
    // disposable properties
    _root : null,           // root path
    _pub : null,            // event publisher
    _ns : null,             // the namespace token
    
    //---------------------------------------------------------------------------
    // Sets the current root path
    root : function root(p) {
        // Get the real path
        try {
            this._root = fs.realpathSync(p);
        }
        catch(e) {
            throw new Error('Unable to resolve the path [' + p + ']. Most likely it does not exist in the file system');
        }
        
        // Extract the base namespace
        this._ns = path.basename(p);
        
        return this;
    },
    
    //---------------------------------------------------------------------------
    // Object maker
    make : function make(obj) {
        var o = Object.create(obj);
        if (o['construct'] === 'function')
            o['construct'].call(o, easy);
        
        return o;
    },
    
    //---------------------------------------------------------------------------
    // Construct an object given the result of the require() function. What this
    // actually does is that it'll fire off an async watcher to wait until the
    // easy namespace has been fully loaded, and once it does it'll fire the
    // object's construct() function to tie it up with other objects in the
    // namespace (mapping dependencies). This ensures that all components that
    // are dependant on each other will get initialized properly at the right
    // time.
    _construct : function construct(o, name) {
        // Make sure o contains the construct function which should be called
        // once all components have been loaded
        if (typeof o['construct'] === 'function') {
            // Bind it to the publisher to watch for the ready signal
            var self = this;
            self._pub.bind('load', function() {
                // Call the component's constructor
                o.construct.call(o, easy);
                console.log('Constructed component [' + name + ']');
                
                // Decrement the loaded counter
                _loaded--;
                
                // Trigger the "ready" event if all components have loaded
                if (_loaded <= 0) {
                    self._pub.unbind('load');
                    self._pub.trigger('ready');
                }
            });
            
            // Increment the loaded counter
            _loaded++;
        }
        
        // Say yeah!
        console.log('Loaded component [' + name + ']');
        
        return o;
    },
    
    //---------------------------------------------------------------------------
    // Ready function to signify that initialization has completed
    //
    // - callback (function)
    //      A callback function to call once initialization is completed
    ready : function ready(callback) {
        var self = this;
        
        // Start walking, Johnny!
        var walker = walk.walk(this._root, { followLinks : false });
        
        // Create a publisher
        this._pub = {};
        _.extend(this._pub, Backbone.Events);

        // Initialize the base namespace
        easy[this._ns] = {};
        
        // Process all files
        walker.on('file', function file(root, stat, next){
            var base = self._ns;
            
            // Remove unnecessary path to create a relative path
            if (root.indexOf(self._root) !== -1)
                root = root.substr(self._root.length);
                
            // Make sure the first token is not a backslash ('/')
            if (root.charAt(0) === '/')
                root = root.substr(1);

            // Split the root by the directory separator
            var fn = path.basename(stat.name, path.extname(stat.name));     // the file name
            var nt = [];                                                    // an array of namespace tokens
            
            if (root) {
                nt = root.split('/');
                root = root + '/';
            }

            // Don't process it if it's "easy.js"!
            if (fn !== 'easy') {
                // Change it to pascal case because the name in the namespace would need to match the filename
                var cn = toPascal(fn);
                
                // Convert nt into a proper associative array
                var ptr = easy[base];                                       // pointer to the current namespace
                var len = nt.length;                                        // length of the namespace tokens array
                for (var i = 0; i < len; i++) {
                    // The current namespace
                    var ns = toCamel(nt[i]);
                    
                    // Don't re-initialize the current namespace it if already exists
                    if (!ptr[ns])
                        ptr[ns] = {};
                        
                    // Move the pointer forward to the next namespace token
                    ptr = ptr[ns];
                }

                // Include the object into the namespace
                var obj = easy._construct(require(self._root + '/' + root + fn), fn);

                // Only add it if the object exists
                if (obj) {
                    ptr[cn] = obj;
                }
            }
            
            // Continue processing the next file
            next();
        });
        
        walker.on('end', function end() {
            
            // LOG: Done loading all components
            console.log((new Date).toUTCString() + ': Done loading all components');
            console.log('--------------------------------------------------------');
            console.log('BEGIN');
            
            // Bind the ready event to be fired after all components have finished
            // their self-construction
            var self = easy;
            self._pub.bind('ready', function go() {
                // Unbind the ready event
                self._pub.unbind('ready');
                
                // Delete the pub property
                delete self._pub;
                
                // Delete the construct function
                delete self._construct;
                
                // Delete the ready function
                delete self.ready;
                
                // Delete the root function
                delete self.root;
                
                // Delete the root property
                delete self._root;
                
                // Delete the namespace property
                delete self._ns;
                
                // Call it nao!!!!
                callback.call(easy);
            });
            
            // Triggers the loading mechanism
            self._pub.trigger('load');
        });
    }
};